@@ -1,11 +1,30 @@
 import base64
 import json
+import time
 from typing import Union, Optional
+import logging
 from urllib3.util import Retry
 from requests.adapters import HTTPAdapter
 
 import requests
 
+logger = logging.getLogger(__name__)
+
+
+def _default_retry_strategy() -> Retry:
+    retry_backoff_factor = 0.7
+    connection_timeout = 5
+    read_timeout = 8
+    method_whitelist = ["GET", "POST"]
+
+    retry_strategy = Retry(
+        backoff_factor=retry_backoff_factor,
+        connect=connection_timeout,
+        read=read_timeout,
+        status_forcelist=[502, 503],
+        allowed_methods=method_whitelist)
+    return retry_strategy
+
 
 class FinalityTypes:
     FINAL = "final"
@@ -13,41 +32,86 @@ class FinalityTypes:
 
 
 class JsonProviderError(Exception):
-    pass
+    def get_type(self) -> Optional[str]:
+        try:
+            return self.args[0]["name"]
+        except (IndexError, KeyError):
+            return None
+
+    def get_cause(self) -> Optional[str]:
+        try:
+            return self.args[0]["cause"]["name"]
+        except (IndexError, KeyError):
+            return None
+
+    def is_invalid_nonce_tx_error(self) -> bool:
+        try:
+            return (
+                self.get_type() == "HANDLER_ERROR"
+                and self.get_cause() == "INVALID_TRANSACTION"
+                and "InvalidNonce" in self.args[0]["data"]["TxExecutionError"]["InvalidTxError"]
+            )
+        except (IndexError, KeyError):
+            return False
+
+    def is_request_timeout_error(self) -> bool:
+        return (
+            self.get_type() == "HANDLER_ERROR"
+            and self.get_cause() == "TIMEOUT_ERROR"
+        )
+
+    def is_expired_tx_error(self) -> bool:
+        try:
+            return (
+                self.get_type() == "HANDLER_ERROR"
+                and self.get_cause() == "INVALID_TRANSACTION"
+                and self.args[0]["data"]["TxExecutionError"]["InvalidTxError"] == "Expired"
+            )
+        except (IndexError, KeyError):
+            return False
 
 
 class JsonProvider(object):
-    def __init__(self, rpc_addr, proxies=None):
+    def __init__(self, rpc_addr, proxies=None,
+                 retry_strategy: Retry = _default_retry_strategy(),
+                 tx_timeout_retry_number: int = 12,
+                 tx_timeout_retry_backoff_factor: float = 1.5,):
         if isinstance(rpc_addr, tuple):
             self._rpc_addr = "http://%s:%s" % rpc_addr
         else:
             self._rpc_addr = rpc_addr
-        self.proxies = proxies
 
-        retry_backoff_factor = 1.0
-        total_retries = 5
-        connection_timeout = 10
-        read_timeout = 12
-        method_whitelist = ["GET", "POST"]
-
-        retry_strategy = Retry(
-            total=total_retries,
-            backoff_factor=retry_backoff_factor,
-            connect=connection_timeout,
-            read=read_timeout,
-            status_forcelist=[502, 503],
-            allowed_methods=method_whitelist)
         adapter = HTTPAdapter(max_retries=retry_strategy)
-
         http = requests.Session()
         http.mount("https://", adapter)
         http.mount("http://", adapter)
+
+        self.proxies = proxies
         self._http = http
 
+        self._tx_timeout_retry_number = tx_timeout_retry_number
+        self._tx_timeout_retry_backoff_factor = tx_timeout_retry_backoff_factor
+
     def rpc_addr(self) -> str:
         return self._rpc_addr
 
-    def json_rpc(self, method: str, params: Union[dict, list, str]) -> dict:
+    def json_rpc(self, method: str, params: Union[dict, list, str], timeout: 'TimeoutType' = 2.0) -> dict:
+        attempt = 0
+        while True:
+            try:
+                return self._json_rpc_once(method, params, timeout)
+            except JsonProviderError as e:
+                if attempt >= self._tx_timeout_retry_number:
+                    raise
+                if e.is_request_timeout_error():
+                    logger.warning(
+                        "Retrying request to %s as it has timed out: %s", method, e)
+                else:
+                    raise
+            attempt += 1
+            time.sleep(self._tx_timeout_retry_backoff_factor ** attempt * 0.5)
+
+    def _json_rpc_once(self, method: str, params: Union[dict, list, str], timeout: 'TimeoutType' = 2.0) -> dict:
         j = {
             'method': method,
             'params': params,
